{*
 * File: ...................... sockschain.pas
 * Autor: ..................... Maxim Artemev aka Bert Raccoon
 * Copyright: ................. (c) 2003 by Max Artemev, MC NTT (www.ntt.ru)
 * Desc: ...................... SOCKS chains support.

 $Id: sockschain.pas,v 1.2 2005/02/15 11:21:21 bert Exp $

 $Log: sockschain.pas,v $
 Revision 1.2  2005/02/15 11:21:21  bert
 *** empty log message ***


}
unit sockschain;

interface
uses Windows, Sysutils, Classes, xml_config, inifiles,
     pinger, winsock2, base64;

type
    TProxyTask = (psNone, ptCheck, ptPing);

    TSocksChainItem = class(TThread)
    private
      FAuth: Boolean;
      FinUse: Boolean;
      FHTTP_Auth: Boolean;
      FPort: integer;
      FVersion: integer;
      FHTTP_Pass: string;
      Fident: string;
      FServer: string;
      FHTTP_User: string;
      FPassword: string;
      FLogin: string;
      FStatus: string;
      FSock: TSocket;
      FOnCheckDone: TNotifyEvent;
      FTask: TProxyTask;
      FData: TObject;
      FOnPingDone: TNotifyEvent;
      FPingMSec: integer;
      function GetProto: string;
      function GetAnon: string;
      function ResolveIP: DWORD;
      function WaitForConnect: integer;
      function DoPing(): integer;
      function DoConnect(Name: TSockAddr): integer;
    public
      constructor CreateIt(); virtual;
      destructor Destroy; override;

      procedure Assign(Source: TObject);

      procedure Execute(); override;
      function Check: string;
      procedure Ping();

      property Protocol: string read GetProto;
      property Anon: string read GetAnon;
      property Status: string read FStatus;

      property Server: string read FServer write FServer;
      property Port: integer read FPort write FPort;
      property Version: integer read FVersion write FVersion;
      property Auth: Boolean read FAuth write FAuth;

      property Login: string read FLogin write FLogin;
      property Password: string read FPassword write FPassword;
      property ident: string read Fident write Fident;
      property HTTP_User: string read FHTTP_User write FHTTP_User;
      property HTTP_Pass: string read FHTTP_Pass write FHTTP_Pass;
      property HTTP_Auth: Boolean read FHTTP_Auth write FHTTP_Auth;
      property inUse: Boolean read FinUse write FinUse;

      property PingMSec: integer read FPingMSec write FPingMSec;

      property Data: TObject read FData write FData;
      property Task: TProxyTask read FTask write FTask;
      property OnCheckDone: TNotifyEvent read FOnCheckDone write FOnCheckDone;
      property OnPingDone: TNotifyEvent read FOnPingDone write FOnPingDone;
    end;

    TSocksChain = class
    private
      FChainList: TList;
      FChainIniFile: string;
      function GetItem(index: integer): TSocksChainItem;
      procedure Clear;
      function GetCount: integer;
      procedure LoadFromOldIni;
    public
      constructor Create(ChainIniFile : string = ''); virtual;
      destructor Destroy; override;

      function GetFirstIndex(): integer;
      function GetNextIndex(FromIndex: integer): integer;
      function GetLastIndex(): integer;

      function AddSocks(serv: string; port: integer; version: integer; login, pass, ident: string; Auth: Boolean; http_user, http_pass: string; http_auth: boolean): TSocksChainItem;
      procedure DelSocks(index: integer);
      procedure LoadFromIni;
      procedure SaveToIni;
      function GetLastSOCKS5(): TSocksChainItem;

      procedure Exchange(index1, index2: integer);

      property Items[index: integer]: TSocksChainItem read GetItem; default;
      property Count: integer read GetCount;

    end;

    procedure Init;
    procedure Fini;

var
   SocksChains: TSocksChain;

implementation
uses cfg, loger, misc;

{ TSocksChain }

function TSocksChain.AddSocks(serv: string; port: integer; version: integer; login, pass, ident: string; Auth: Boolean; http_user, http_pass: string; http_auth: boolean): TSocksChainItem;
var
   Item: TSocksChainItem;
begin
     Item := TSocksChainItem.CreateIt();
     Item.Server := serv;
     Item.Port := port;
     Item.Version := version;
     Item.Login := login;
     Item.Password := pass;
     Item.Auth := auth;
     Item.ident := ident;
     Item.HTTP_User := http_user;
     Item.HTTP_Pass := http_pass;
     Item.HTTP_Auth := http_auth;

     FChainList.Add(Item);
     result := Item;
end;

constructor TSocksChain.Create(ChainIniFile : string { = '' });
begin
     FChainList := TList.Create;
     FChainIniFile := ChainIniFile;
     LoadFromIni;
end;

procedure TSocksChain.Clear;
var
   i: integer;
   Item: TSocksChainItem;
begin
     for i := FChainList.Count - 1 downto 0 do
     begin
          Item := TSocksChainItem(FChainList[i]);
          Item.Terminate();
          Item.Free;
     end;
     FChainList.CLear;
end;

function TSocksChain.GetCount: integer;
begin
     result := FChainList.Count;
end;


procedure TSocksChain.DelSocks(index: integer);
begin
     TSocksChainItem(FChainList[index]).Terminate();
     TSocksChainItem(FChainList[index]).Free;
     FChainList.Delete(index);
end;


destructor TSocksChain.Destroy;
begin
     SaveToIni;
     Clear();
     FChainList.Free;
     FChainList := nil;
     inherited Destroy;
end;


function TSocksChain.GetItem(index: integer): TSocksChainItem;
begin
     result := nil;
     if (index >= 0) and (index < FChainList.Count) then
        result := TSocksChainItem(FChainList[index]);
end;

procedure TSocksChain.LoadFromOldIni;
var
   ini: TIniFile;
   Sections, Values: TStringList;
   i: integer;
   Item: TSocksChainItem;
begin
     Clear;
     Sections := TStringList.Create;
     Values   := TStringList.Create;

     if FChainIniFile = '' then
        Ini := TIniFile.Create('freechain.ini')
     else
        Ini := TIniFile.Create(FChainIniFile);

     Ini.ReadSections(Sections);
     for i:=0 to Sections.Count - 1 do
     begin
          Values.Clear;
          Ini.ReadSectionValues(Sections[i], Values);
          Item := TSocksChainItem.CreateIt();
          Item.Server := Sections[i];
          Item.Port := StrToIntDef(Values.Values['Port'], 1080);
          Item.Version := StrToIntDef(Values.Values['Version'], PROXY_VER_SOCKS5);
          Item.Login := Values.Values['Login'];
          Item.ident := Values.Values['Socks4Ident'];
          Item.Password := Values.Values['Password'];

          Item.HTTP_User := Values.Values['HttpUser'];
          Item.HTTP_Pass := Values.Values['HttpPass'];
          Item.HTTP_Auth := Boolean(StrToIntDef(Values.Values['HttpAuth'], 0));
          Item.inUse := Boolean(StrToIntDef(Values.Values['InUse'], 1));

          Item.auth := Boolean(StrToIntDef(Values.Values['Auth'], 0));
          FChainList.Add(Item);
     end;
     Values.Free;
     Sections.Free;
     ini.Free;
end;



procedure TSocksChain.LoadFromIni;
var
   ini: TXMLConfig;
   Sections, Values: TStringList;
   i: integer;
   Item: TSocksChainItem;
begin
     if isConfigExists('freechain.ini') and (not isConfigExists('freecap.xml')) then
     begin
          LoadFromOldIni();
          exit;
     end;

     Clear;
     Sections := TStringList.Create;
     Values   := TStringList.Create;

     if FChainIniFile = '' then
        Ini := TXMLConfig.Create('freecap.xml')
     else
        Ini := TXMLConfig.Create(FChainIniFile);

     Ini.ReadSections(PART_SOCKSCHAIN, Sections);
     for i:=0 to Sections.Count - 1 do
     begin
          Values.Clear;
          Ini.ReadSectionValues(PART_SOCKSCHAIN, Sections[i], Values);
          Item := TSocksChainItem.CreateIt();
          Item.Server := Sections[i];

          if Values.Values['Server'] <> '' then
            Item.Server := Values.Values['Server'];


          Item.Port := StrToIntDef(Values.Values['Port'], 1080);
          Item.Version := StrToIntDef(Values.Values['Version'], PROXY_VER_SOCKS5);
          Item.Login := Values.Values['Login'];
          Item.ident := Values.Values['Socks4Ident'];
          Item.Password := Values.Values['Password'];

          Item.HTTP_User := Values.Values['HttpUser'];
          Item.HTTP_Pass := Values.Values['HttpPass'];
          Item.HTTP_Auth := Boolean(StrToIntDef(Values.Values['HttpAuth'], 0));
          Item.inUse := Boolean(StrToIntDef(Values.Values['InUse'], 1));

          Item.Auth := Boolean(StrToIntDef(Values.Values['Auth'], 0));
          FChainList.Add(Item);
     end;
     Values.Free;
     Sections.Free;
     ini.Free;
end;

procedure TSocksChain.SaveToIni;
var
   ini: TXMLConfig;
   Sections: TStringList;
   i  : integer;
   Item: TSocksChainItem;
   ident: string;
begin
     Sections := TStringList.Create;

     if FChainIniFile = '' then
        Ini := TXMLConfig.Create('freecap.xml')
     else
        Ini := TXMLConfig.Create(FChainIniFile);

     Ini.ReadSections(PART_SOCKSCHAIN, Sections);
     for i:=0 to Sections.Count - 1 do
       ini.EraseSection(PART_SOCKSCHAIN, Sections[i]);

     for i:=0 to FChainList.Count - 1 do
     begin
          Item := TSocksChainItem(FChainList[i]);
          Ident := IntToStr(i);

          ini.WriteString(PART_SOCKSCHAIN, Ident, 'Server', Item.Server);
          ini.WriteInteger(PART_SOCKSCHAIN, Ident, 'Port', Item.Port);
          ini.WriteInteger(PART_SOCKSCHAIN, Ident, 'Version', Item.Version);
          ini.WriteString(PART_SOCKSCHAIN, Ident, 'Login', Item.Login);
          ini.WriteString(PART_SOCKSCHAIN, Ident, 'Socks4Ident', Item.ident);
          ini.WriteString(PART_SOCKSCHAIN, Ident, 'Password', Item.Password);
          ini.WriteBool(PART_SOCKSCHAIN, Ident, 'Auth', Item.Auth);
          ini.WriteString(PART_SOCKSCHAIN, Ident, 'HttpUser', Item.HTTP_User);
          ini.WriteString(PART_SOCKSCHAIN, Ident, 'HttpPass', Item.HTTP_Pass);
          ini.WriteBool(PART_SOCKSCHAIN, Ident, 'HttpAuth', Item.HTTP_Auth);
          ini.WriteBool(PART_SOCKSCHAIN, Ident, 'InUse', Item.inUse);
     end;

     ini.Free;
     Sections.Free;
end;

procedure TSocksChain.Exchange(index1, index2: integer);
var
  Item: TSocksChainItem;
begin
     Item := TSocksChainItem.CreateIt();
     Item.Assign(TSocksChainItem(FChainList[Index1]));
     TSocksChainItem(FChainList[Index1]).Assign(TSocksChainItem(FChainList[Index2]));
     TSocksChainItem(FChainList[Index2]).Assign(Item);
     Item.Free;
end;


function TSocksChain.GetFirstIndex(): integer;
var
   i: integer;
begin
     result := -1;
     if Count = 0 then exit;

     i := 0;
     while (i < Count) and (not SocksChains[i].inUse) do
       inc(i);

     if (i <> Count) then
        result := i;
end;


function TSocksChain.GetNextIndex(FromIndex: integer): integer;
begin
     repeat
          inc(FromIndex);
     until (FromIndex = Count) or (SocksChains[FromIndex].inUse);
     result := FromIndex;
end;

function TSocksChain.GetLastIndex: integer;
begin
     result := Count - 1;
     if result = 0 then
     begin
          if not SocksChains[result].inUse then
            result := -1;
     end
     else
     begin
          while (result >= 0) and (not SocksChains[result].inUse) do
            dec(result);
     end;
end;


function TSocksChain.GetLastSOCKS5: TSocksChainItem;
var
   i: integer;
begin
     result := nil;
     i := Count - 1;

     if (i = 0) then
     begin
          if (SocksChains[i].inUse) and (SocksChains[i].Version = 5) then
            result := SocksChains[i];
     end
     else
     begin
          while (i >= 0) and ((not SocksChains[i].inUse) or (SocksChains[i].Version <> 5)) do
            dec(i);
          if i = -1 then
          begin
               Log(LOG_LEVEL_WARN, 'GetLastSOCKS5 = nil', []);
               exit;
          end;
          result := SocksChains[i];
     end;
end;

{ TSocksChainItem }

constructor TSocksChainItem.CreateIt;
begin
     inherited Create(True);
     FTask := psNone;
     FreeOnTerminate := False;
end;


function TSocksChainItem.GetProto: string;
begin
     case FVersion of
       1: result := 'http';
       4: result := 'socks4';
       5: result := 'socks5';
     else
        result := 'Unknown';
     end;
end;


function TSocksChainItem.GetAnon: string;
begin
     result := 'Y';
     if FAuth or FHTTP_Auth then
        result := 'N'
end;



function TSocksChainItem.DoPing: integer;
begin
     result := pinger.Ping(FServer);
end;

function TSocksChainItem.ResolveIP: DWORD;
var
   HostEntry: PHostEnt;
   ph: PHostAddr;
begin
     result := inet_addr(PChar(FServer));
     if result <> INADDR_NONE then
        exit;

     result := 0;
     HostEntry := gethostbyname(PChar(FServer));
     if HostEntry = nil then
       exit;
     ph := PHostAddr(HostEntry^.h_addr_list^);
     if ph <> nil then
       result := Cardinal(ph^);
end;

function TSocksChainItem.Check: string;
begin
     FTask := ptCheck;
     Resume;
end;

procedure TSocksChainItem.Ping;
begin
     FTask := ptPing;
     Resume;
end;


function TSocksChainItem.DoConnect(Name: TSockAddr): integer;
begin
     connect(FSock, @Name, SizeOf(name));
     result := WaitForConnect();
end;

function TSocksChainItem.WaitForConnect: integer;
var
   res     : integer;
   writefd : TFDSet;
   errorfd : TFDSet;
   timeval : TTimeVal;
begin
     {* Wait until socket will be connected
      *}
     res := 1;
     timeval.tv_sec := 1;
     timeval.tv_usec := 0;

     ZeroMemory(@writefd, SizeOf(writefd));
     ZeroMemory(@errorfd, SizeOf(errorfd));

     while (not Terminated) do
     begin
          FD_CLR(FSock, writefd);
          FD_SET(FSock, writefd);

          FD_CLR(FSock, errorfd);
          FD_SET(FSock, errorfd);

          res := select(0, nil, @writefd, @errorfd, @timeval);

          if FD_ISSET(FSock, writefd) then
          begin
               result := 0;
               exit;
          end;

          if FD_ISSET(FSock, errorfd) or (res < 0) then
          begin
               FStatus := 'Connection refused';
               break;
          end;
     end;
     result := res;
end;


procedure TSocksChainItem.Execute();
var
   Name: TSockAddr;
   res: integer;
   arg: DWORD;
   AvgPing: integer;
begin
     while not Terminated do
     begin
          if FTask = ptCheck then
          begin
               AvgPing := Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing div 4;
               FPingMSec := AvgPing;
               if Assigned(FOnPingDone) then
                 FOnPingDone(Self);

               Name.sin_family := AF_INET;
               Name.sin_addr.S_addr := ResolveIP();
               Name.sin_port := ntohs(port);

               if (Name.sin_addr.S_addr = 0) then
               begin
                    FStatus := 'Host not found';
               end
               else
               begin
                    FSock := socket(AF_INET, SOCK_STREAM, 0);
                    arg := 1;
                    ioctlsocket(FSock, FIONBIO, arg);
                    res := Self.DoConnect(Name);
                    if (res = 0) then
                      FStatus := 'OK';

                    closesocket(FSock);
               end;
               if Assigned(FOnCheckDone) then
                 FOnCheckDone(Self);
               Suspend;
          end;

          if FTask = ptPing then
          begin
               AvgPing := Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing + Self.DoPing();
               AvgPing := AvgPing div 4;
               FPingMSec := AvgPing;
               if Assigned(FOnPingDone) then
                 FOnPingDone(Self);
               Suspend;
          end;
     end;
end;



procedure TSocksChainItem.Assign(Source: TObject);
var
   Src: TSocksChainItem;
begin
     if Source is TSocksChainItem then
     begin
          Src := (Source as TSocksChainItem);
          FServer := Src.Server;
          FPort :=  Src.Port;
          FVersion := Src.Version;
          FAuth := Src.Auth;
          FLogin := Src.Login;
          FPassword := Src.Password;
          Fident := Src.ident;
          FHTTP_User := Src.HTTP_User;
          FHTTP_Pass := Src.HTTP_Pass;
          FHTTP_Auth := Src.HTTP_Auth;
          FinUse := Src.inUse;
          FData := Src.Data;
     end;
end;


procedure Init;
var
   Buf: array[0..MAX_PATH] of Char;
begin
     ZeroMemory(@Buf, SizeOf(Buf));
     if GetEnvironmentVariable('FreeCAPConfigFile', @Buf, SizeOf(Buf)) <> 0 then
       SocksChains := TSocksChain.Create(Buf)
     else
       SocksChains := TSocksChain.Create;
end;


procedure Fini;
begin
     SocksChains.Clear();
     SocksChains.Free;
end;


destructor TSocksChainItem.Destroy;
begin
     inherited Destroy;
end;

end.
